# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from amazon_kclpy import messages


class MalformedAction(Exception):
    """
    Raised when an action given by the MultiLangDaemon doesn't have all the appropriate attributes.
    """
    pass


_serializers = {
    "initialize": messages.InitializeInput,
    "processRecords": messages.ProcessRecordsInput,
    "shutdown": messages.ShutdownInput,
    "checkpoint": messages.CheckpointInput,
    "record": messages.Record,
    "shutdownRequested": messages.ShutdownRequestedInput,
    "leaseLost": messages.LeaseLostInput,
    "shardEnded": messages.ShardEndedInput,
}


def _format_serializer_names():
    return ", ".join('"{k}"'.format(k=k) for k in _serializers.keys())


def message_decode(json_dict):
    """
    Translates incoming JSON commands into MessageDispatch classes

    :param dict json_dict: a dictionary of JSON data

    :return: an object that can be used to dispatch the received JSON command
    :rtype: amazon_kclpy.messages.MessageDispatcher

    :raises MalformedAction: if the JSON object is missing action, or an appropriate serializer for that
        action can't be found
    """
    try:
        action = json_dict["action"]
    except KeyError as key_error:
        raise MalformedAction("Action {json_dict} was expected to have key {key!s}".format(json_dict=json_dict,
                                                                                           key=key_error))
    try:
        serializer = _serializers[action]
    except KeyError:
        raise MalformedAction("Received an action which couldn't be understood. Action was '{action}' -- Allowed {keys}"
                              .format(action=action, keys=_format_serializer_names()))

    return serializer(json_dict)



# Copyright 2014-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import abc


class RecordProcessorBase(object):
    """
    Base class for implementing a record processor.A RecordProcessor processes a shard in a stream.
    Its methods will be called with this pattern:

    - initialize will be called once
    - process_records will be called zero or more times
    - shutdown will be called if this MultiLangDaemon instance loses the lease to this shard
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def initialize(self, initialize_input):
        """
        Called once by a KCLProcess before any calls to process_records

        :param amazon_kclpy.messages.InitializeInput initialize_input: Information about the
            initialization request for the record processor
        """
        raise NotImplementedError

    @abc.abstractmethod
    def process_records(self, process_records_input):
        """
        Called by a KCLProcess with a list of records to be processed and a checkpointer which accepts sequence numbers
        from the records to indicate where in the stream to checkpoint.

        :param amazon_kclpy.messages.ProcessRecordsInput process_records_input: the records, and metadata about the
            records.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown(self, shutdown_input):
        """
        Called by a KCLProcess instance to indicate that this record processor should shutdown. After this is called,
        there will be no more calls to any other methods of this record processor.

        As part of the shutdown process you must inspect :attr:`amazon_kclpy.messages.ShutdownInput.reason` to
        determine the steps to take.

            * Shutdown Reason ZOMBIE:
                **ATTEMPTING TO CHECKPOINT ONCE A LEASE IS LOST WILL FAIL**

                A record processor will be shutdown if it loses its lease.  In this case the KCL will terminate the
                record processor.  It is not possible to checkpoint once a record processor has lost its lease.
            * Shutdown Reason TERMINATE:
                **THE RECORD PROCESSOR MUST CHECKPOINT OR THE KCL WILL BE UNABLE TO PROGRESS**

                A record processor will be shutdown once it reaches the end of a shard.  A shard ending indicates that
                it has been either split into multiple shards or merged with another shard.  To begin processing the new
                shard(s) it's required that a final checkpoint occurs.


        :param amazon_kclpy.messages.ShutdownInput shutdown_input: Information related to the shutdown request
        """
        raise NotImplementedError

    def shutdown_requested(self, shutdown_requested_input):
        """
        Called by a KCLProcess instance to indicate that this record processor is about to be be shutdown.  This gives
        the record processor a chance to checkpoint, before the lease is terminated.

        :param amazon_kclpy.messages.ShutdownRequestedInput shutdown_requested_input:
            Information related to shutdown requested.
        """
        pass

    version = 2


class V1toV2Processor(RecordProcessorBase):
    """
    Provides a bridge between the new v2 RecordProcessorBase, and the original RecordProcessorBase.

    This handles the conversion of the new input types to the older expected forms.  This normally shouldn't be used
    directly by record processors, since it's just a compatibility layer.

    The delegate should be a :py:class:`amazon_kclpy.kcl.RecordProcessorBase`:

    """
    def __init__(self, delegate):
        """
        Creates a new V1 to V2 record processor.

        :param amazon_kclpy.kcl.RecordProcessorBase delegate: the delegate where requests will be forwarded to
        """
        self.delegate = delegate

    def initialize(self, initialize_input):
        """
        Initializes the record processor

        :param amazon_kclpy.messages.InitializeInput initialize_input: the initialization request
        :return: None
        """
        self.delegate.initialize(initialize_input.shard_id)

    def process_records(self, process_records_input):
        """
        Expands the requests, and hands it off to the delegate for processing

        :param amazon_kclpy.messages.ProcessRecordsInput process_records_input: information about the records
            to process
        :return: None
        """
        self.delegate.process_records(process_records_input.records, process_records_input.checkpointer)

    def shutdown(self, shutdown_input):
        """
        Sends the shutdown request to the delegate

        :param amazon_kclpy.messages.ShutdownInput shutdown_input: information related to the record processor shutdown
        :return: None
        """
        self.delegate.shutdown(shutdown_input.checkpointer, shutdown_input.reason)

    def shutdown_requested(self, shutdown_requested_input):
        """
        Sends the shutdown request to the delegate

        :param amazon_kclpy.messages.ShutdownInput shutdown_input: information related to the record processor shutdown
        :return: None
        """
        self.delegate.shutdown_requested(shutdown_requested_input.checkpointer)



# Copyright 2014-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import abc
import base64
from datetime import datetime

from amazon_kclpy.checkpoint_error import CheckpointError


class MessageDispatcher(object):
    """
    The base class use to dispatch actions to record processors.  This allows derived classes
    to determine which method on the record processor they need to call.  Additionally classes
    implementing this generally wrap up the parameters into themselves
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def dispatch(self, checkpointer, record_processor):
        """
        Dispatches the current message to the record processor.

        :param amazon_kclpy.kcl.Checkpointer checkpointer: The checkpointer that can be used by the record
            process to record its progress

        :param amazon_kclpy.v3.processor.RecordProcessorBase record_processor: The record processor that will receive,
            and process the message.

        :return: Nothing
        """
        raise NotImplementedError

    @abc.abstractmethod
    def action(self):
        """
        Retrieves the name of the action that caused the creation of this dispatcher.

        :return str: The name of the action e.g. initialize, or processRecords
        """
        raise NotImplementedError


class InitializeInput(MessageDispatcher):
    """
    Provides the necessary parameters to initialize a Record Processor
    """
    def __init__(self, json_dict):
        """
        Configures the input, preparing it for dispatch

        :param dict json_dict: The raw representation of the JSON data
        """
        self._shard_id = json_dict["shardId"]
        self._sequence_number = json_dict["sequenceNumber"]
        self._sub_sequence_number = json_dict["subSequenceNumber"]
        self._action = json_dict['action']

    @property
    def shard_id(self):
        """
        The shard id that this record processor will be processing.

        :return: the shard id
        :rtype: str
        """
        return self._shard_id

    @property
    def sequence_number(self):
        """
        The sequence number that this record processor will start at.  This can be None if this record processor is
        starting on a fresh shard.

        :return: the sequence number
        :rtype: str or None
        """
        return self._sequence_number

    @property
    def sub_sequence_number(self):
        """
        The sub sequence number that this record processor will start at.  This will never be none,
        but can be 0 if there was no sub-sequence number

        :return: the subsequence number
        :rtype: int
        """
        return self._sub_sequence_number

    @property
    def action(self):
        """
        The action that spawned this message

        :return: the original action value
        :rtype: str
        """
        return self._action

    def dispatch(self, checkpointer, record_processor):
        record_processor.initialize(self)


class ProcessRecordsInput(MessageDispatcher):
    """
    Provides the records, and associated metadata for calls to process_records.
    """
    def __init__(self, json_dict):
        self._records = json_dict["records"]
        self._millis_behind_latest = json_dict["millisBehindLatest"]
        self._checkpointer = None
        self._action = json_dict['action']

    @property
    def records(self):
        """
        The records that are part of this request.

        :return: records that are part of this request
        :rtype: list[amazon_kclpy.messages.Record]
        """
        return self._records

    @property
    def millis_behind_latest(self):
        """
        An approximation of how far behind the current time this batch of records is.

        :return: the number of milliseconds
        :rtype: int
        """
        return self._millis_behind_latest

    @property
    def checkpointer(self):
        """
        Provides the checkpointer that will confirm all records upto, and including this batch of records.

        :return: the checkpointer for this request
        :rtype: amazon_kclpy.kcl.Checkpointer
        """
        return self._checkpointer

    @property
    def action(self):
        """
        The action that spawned this message

        :return: the original action value
        :rtype: str
        """
        return self._action

    def dispatch(self, checkpointer, record_processor):
        self._checkpointer = checkpointer
        record_processor.process_records(self)


class LeaseLostCheckpointer:
    """
    Checkpointer for use in the lease loss scenario.  This is specifically used in the v2 to v3 delegate
    """
    def __init__(self):
        pass

    def checkpoint(self, sequence_number=None, sub_sequence_number=None):
        raise CheckpointError("Checkpointing is not allowed when the lease is lost")


class ShutdownInput:
    """
    Used to tell the record processor it will be shutdown.
    """

    @staticmethod
    def zombie():
        return ShutdownInput("ZOMBIE", LeaseLostCheckpointer())

    @staticmethod
    def terminate(checkpointer):
        return ShutdownInput("TERMINATE", checkpointer)

    def __init__(self, reason, checkpointer):
        self._action = 'shutdown'
        self._reason = reason
        self._checkpointer = checkpointer

    @property
    def reason(self):
        """
        The reason that this record processor is being shutdown, will be one of

        * TERMINATE
        * ZOMBIE

        :return: the reason for the shutdown
        :rtype: str
        """
        return self._reason

    @property
    def checkpointer(self):
        """
        The checkpointer that can be used to checkpoint this shutdown.

        :return: the checkpointer
        :rtype: amazon_kclpy.kcl.Checkpointer
        """
        return self._checkpointer

    @property
    def action(self):
        """
        The action that spawned this message

        :return: the original action value
        :rtype: str
        """
        return self._action


class LeaseLostInput(MessageDispatcher):
    """
    Message, and input that is sent when the client has lost the lease for this shard.
    """

    def __init__(self, json_dict):
        self._action = json_dict['action']

    @property
    def action(self):
        """
        The action that indicated the lease was lost

        :return: the name of the action
        :rtype: str
        """
        return self._action

    def dispatch(self, checkpointer, record_processor):
        """
        Dispatch the lease lost notification to the record processor

        :param checkpointer: unused
        :param record_processor: the record processor to dispatch the call to
        :return: None
        """
        record_processor.lease_lost(self)


class ShardEndedInput(MessageDispatcher):
    """
    Message and input that is sent to the record processor when the client has reached the end of the shard.
    """

    def __init__(self, json_dict):
        self._checkpointer = None
        self._action = json_dict['action']

    @property
    def action(self):
        """
        The action that caused the creation of this message

        :return: the action name
        :rtype: str
        """
        return self._action

    @property
    def checkpointer(self):
        """
        The checkpointer that the record processor will use to checkpoint the end of the shard
        :return: the checkpointer
        :rtype: Checkpointer
        """
        return self._checkpointer

    def dispatch(self, checkpointer, record_processor):
        """
        Dispatches the shard ended message to the record processor

        :param checkpointer: the checkpointer to be used to officially end processing on the shard
        :param record_processor: the record processor that will handle the shard end message
        """
        self._checkpointer = checkpointer
        record_processor.shard_ended(self)


class ShutdownRequestedInput(MessageDispatcher):
    """
    Used to tell the record processor it will be shutdown.
    """
    def __init__(self, json_dict):
        self._checkpointer = None
        self._action = json_dict['action']

    @property
    def checkpointer(self):
        """
        The checkpointer that can be used to checkpoint before actual shutdown.

        :return: the checkpointer
        :rtype: amazon_kclpy.kcl.Checkpointer
        """
        return self._checkpointer

    @property
    def action(self):
        """
        The action that spawned this message

        :return: the original action value
        :rtype: str
        """
        return self._action

    def dispatch(self, checkpointer, record_processor):
        self._checkpointer = checkpointer
        record_processor.shutdown_requested(self)


class CheckpointInput(object):
    """
    Used in preparing the response back during the checkpoint process.  This shouldn't be used by record processors.
    """
    def __init__(self, json_dict):
        """
        Creates a new CheckpointInput object with the given sequence number, and sub-sequence number.
        The provided dictionary must contain:
        * sequenceNumber
        * subSequenceNumber

        The provided dictionary can optionally contain:
        * error

        :param dict json_dict:
        """
        self._sequence_number = json_dict["sequenceNumber"]
        self._sub_sequence_number = json_dict["subSequenceNumber"]
        self._error = json_dict.get("error", None)

    @property
    def sequence_number(self):
        """
        The sequence number that record processor intends to checkpoint at.  Can be None if the default 
        checkpoint behavior is desired.

        :return: the sequence number
        :rtype: str or None
        """
        return self._sequence_number

    @property
    def sub_sequence_number(self):
        """
        The sub-sequence number that the record processor intends to checkpoint at.  Can be None if 
        the default checkpoint behavior is desired.

        :return: the sub-sequence number
        :rtype: int or None
        """
        return self._sub_sequence_number

    @property
    def error(self):
        """
        The error message that may have resulted from checkpointing.  This will be None if no error occurred.

        :return: the error message
        :rtype: str or None
        """
        return self._error


class Record(object):
    """
    Represents a single record as returned by Kinesis, or Disaggregated from the Kinesis Producer Library
    """
    def __init__(self, json_dict):
        """
        Creates a new Record object that represent a single record in Kinesis.  Construction for the provided
        dictionary requires that the following fields are present:
            * sequenceNumber
            * subSequenceNumber
            * approximateArrivalTimestamp
            * partitionKey
            * data

        :param dict json_dict:
        """
        self._sequence_number = json_dict["sequenceNumber"]
        self._sub_sequence_number = json_dict["subSequenceNumber"]

        self._timestamp_millis = int(json_dict["approximateArrivalTimestamp"])
        self._approximate_arrival_timestamp = datetime.fromtimestamp(self._timestamp_millis / 1000.0)

        self._partition_key = json_dict["partitionKey"]
        self._data = json_dict["data"]
        self._json_dict = json_dict

    @property
    def binary_data(self):
        """
        The raw binary data automatically decoded from the Base 64 representation provided by

        :py:attr:`data` the original source of the data

        :return: a string representing the raw bytes from
        :rtype: str
        """
        return base64.b64decode(self._data)
    
    @property
    def sequence_number(self):
        """
        The sequence number for this record.  This number maybe the same for other records, if they're
        all part of an aggregated record.  In that case the sub_sequence_number will be greater than 0

        :py:attr:`sub_sequence_number`

        :return: the sequence number
        :rtype: str
        """
        return self._sequence_number
    
    @property
    def sub_sequence_number(self):
        """
        The sub-sequence number of this record.  This is only populated when the record is a disaggregated
        record produced by the `amazon-kinesis-producer-library <https://github.com/awslabs/amazon-kinesis-producer>`

        :return: the sub-sequence number
        :rtype: int
        """
        return self._sub_sequence_number

    @property
    def timestamp_millis(self):
        """
        The timestamp of the approximate arrival time of the record in milliseconds since the Unix epoch

        :return: the timestamp in milliseconds
        :rtype: int
        """
        return self._timestamp_millis

    @property
    def approximate_arrival_timestamp(self):
        """
        The approximate time when this record was accepted, and stored by Kinesis.

        :return: the timestamp
        :rtype: datetime
        """
        return self._approximate_arrival_timestamp

    @property
    def partition_key(self):
        """
        The partition key for this record

        :return: the partition key
        :rtype: str
        """
        return self._partition_key

    @property
    def data(self):
        """
        The Base64 encoded data of this record.

        :return: a string containing the Base64 data
        :rtype: str
        """
        return self._data

    def get(self, field):
        return self._json_dict[field]
    
    def __getitem__(self, field):
        return self.get(field)




#!/usr/bin/env python
# Copyright 2014-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This script provides two utility functions:

    ``--print_classpath``
        which prints a java class path. It optionally takes --properties
        and any number of --path options. It will generate a java class path which will include
        the properties file and paths and the location of the KCL jars based on the location of
        the amazon_kclpy.kcl module.

    ``--print_command``
        which prints a command to run an Amazon KCLpy application. It requires a --java
        and --properties argument and optionally takes any number of --path arguments to prepend
        to the classpath that it generates for the command.
"""
from __future__ import print_function
from amazon_kclpy import kcl
from glob import glob
import os
import argparse
import sys
import samples




def get_dir_of_file(f):
    '''
    Returns the absolute path to the directory containing the specified file.

    :type f: str
    :param f: A path to a file, either absolute or relative

    :rtype:  str
    :return: The absolute path of the directory represented by the relative path provided.
    '''
    return os.path.dirname(os.path.abspath(f))

def get_kcl_dir():
    '''
    Returns the absolute path to the dir containing the amazon_kclpy.kcl module.

    :rtype: str
    :return: The absolute path of the KCL package. 
    '''
    return get_dir_of_file(kcl.__file__)

def get_kcl_jar_path():
    '''
    Returns the absolute path to the KCL jars needed to run an Amazon KCLpy app.

    :rtype: str
    :return: The absolute path of the KCL jar files needed to run the MultiLangDaemon.
    '''
    return ':'.join(glob(os.path.join(get_kcl_dir(), 'jars', '*jar')))

def get_kcl_classpath(properties=None, paths=[]):
    '''
    Generates a classpath that includes the location of the kcl jars, the
    properties file and the optional paths.

    :type properties: str
    :param properties: Path to properties file.

    :type paths: list
    :param paths: List of strings. The paths that will be prepended to the classpath.

    :rtype: str
    :return: A java class path that will allow your properties to be found and the MultiLangDaemon and its deps and
        any custom paths you provided.
    '''
    # First make all the user provided paths absolute
    paths = [os.path.abspath(p) for p in paths]
    # We add our paths after the user provided paths because this permits users to
    # potentially inject stuff before our paths (otherwise our stuff would always
    # take precedence).
    paths.append(get_kcl_jar_path())
    if properties:
        # Add the dir that the props file is in
        dir_of_file = get_dir_of_file(properties)
        paths.append(dir_of_file)
    return ":".join([p for p in paths if p != ''])

def get_kcl_app_command(args, multi_lang_daemon_class, properties, log_configuration, paths=[]):
    '''
    Generates a command to run the MultiLangDaemon.

    :type java: str
    :param java: Path to java

    :type multi_lang_daemon_class: str
    :param multi_lang_daemon_class: Name of multi language daemon class e.g. com.amazonaws.services.kinesis.multilang.MultiLangDaemon

    :type properties: str
    :param properties: Optional properties file to be included in the classpath.

    :type paths: list
    :param paths: List of strings. Additional paths to prepend to the classpath.

    :rtype: str
    :return: A command that will run the MultiLangDaemon with your properties and custom paths and java.
    '''
    return "{java} -cp {cp} {daemon} {props} {log_config}".format(java=args.java,
                                    cp = get_kcl_classpath(args.properties, paths),
                                    daemon = multi_lang_daemon_class,
                                    # Just need the basename because the path is added to the classpath
                                    props = properties,
                                    log_config = log_configuration)

if __name__ == '__main__':
    parser = argparse.ArgumentParser("A script for generating a command to run an Amazon KCLpy app")
    parser.add_argument("--print_classpath", dest="print_classpath", action="store_true",
                        default=False,
                        help="Print a java class path.\noptional arguments: --path")
    parser.add_argument("--print_command", dest="print_command", action="store_true",
                        default=False,
                        help="Print a command for running an Amazon KCLpy app.\nrequired "
                        + "args: --java --properties\noptional args: --classpath")
    parser.add_argument("-j", "--java", dest="java",
                        help="The path to the java executable e.g. <some root>/jdk/bin/java",
                        metavar="PATH_TO_JAVA")
    parser.add_argument("-p", "--properties", "--props", "--prop", dest="properties",
                        help="The path to a properties file (relative to where you are running this script)",
                        metavar="PATH_TO_PROPERTIES")
    parser.add_argument("--sample", "--sample-props", "--use-sample-properties", dest="use_sample_props",
                        help="This will use the sample.properties file included in this package as the properties file.",
                        action="store_true", default=False)
    parser.add_argument("-c", "--classpath", "--path", dest="paths", action="append", default=[],
                        help="Additional path to add to java class path. May be specified any number of times",
                        metavar="PATH")
    parser.add_argument("-l", "--log-configuration", dest="log_configuration",
                        help="This will use the logback.xml which will be used by the KCL to log.",
                        metavar="PATH_TO_LOG_CONFIGURATION")
    args = parser.parse_args()
    # Possibly replace the properties with the sample. Useful if they just want to run the sample app.
    if args.use_sample_props:
        if args.properties:
            sys.stderr.write('Replacing provided properties with sample properties due to arg --sample\n')
        args.properties = os.path.join(get_dir_of_file(samples.__file__), 'sample.properties')

    # Print what the asked for
    if args.print_classpath:
        print(get_kcl_classpath(args.properties, args.paths))
    elif args.print_command:
        if args.java and args.properties:
            multi_lang_daemon_class = 'software.amazon.kinesis.multilang.MultiLangDaemon'
            properties_argument = "--properties-file {props}".format(props = args.properties)
            log_argument = ''
            if args.log_configuration is not None:
                log_argument = "--log-configuration {log}".format(log = args.log_configuration)
            print(get_kcl_app_command(args, multi_lang_daemon_class, properties_argument, log_argument, paths=args.paths))
        else:
            sys.stderr.write("Must provide arguments: --java and --properties\n")
            parser.print_usage()
    else:
        parser.print_usage()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc

from amazon_kclpy.messages import ShutdownInput


class RecordProcessorBase(object):
    """
    Base class for implementing a record processor. Each RecordProcessor processes a single shard in a stream.

    The record processor represents a lifecycle where it will be initialized, possibly process records, and
    finally be terminated.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def initialize(self, initialize_input):
        """
        Called once by a the KCL to allow the record processor to configure itself before starting to process records.

        :param amazon_kclpy.messages.InitializeInput initialize_input: Information about the
            initialization request for the record processor
        """
        raise NotImplementedError

    @abc.abstractmethod
    def process_records(self, process_records_input):
        """
        This is called whenever records are received.  The method will be provided the batch of records that were
        received.  A checkpointer is also supplied that allows the application to checkpoint its progress within the
        shard.

        :param amazon_kclpy.messages.ProcessRecordsInput process_records_input: the records, metadata about the
            records, and a checkpointer.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def lease_lost(self, lease_lost_input):
        """
        This is called whenever the record processor has lost the lease.  After this returns the record processor will
        be shutdown.  Additionally once a lease has been lost checkpointing is no longer possible.

        :param amazon_kclpy.messages.LeaseLostInput lease_lost_input: information about the lease loss (currently empty)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def shard_ended(self, shard_ended_input):
        """
        This is called whenever the record processor has reached the end of the shard. The record processor needs to
        checkpoint to notify the KCL that it's ok to start processing the child shard(s).  Failing to checkpoint will
        trigger a retry of the shard end

        :param amazon_kclpy.messages.ShardEndedInput shard_ended_input: information about reaching the end of the shard.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown_requested(self, shutdown_requested_input):
        """
        Called when the parent process is preparing to shutdown.  This gives the record processor one more chance to
        checkpoint before its lease will be released.

        :param amazon_kclpy.messages.ShutdownRequestedInput shutdown_requested_input:
            Information related to shutdown requested including the checkpointer.
        """
        raise NotImplementedError

    version = 3


class V2toV3Processor(RecordProcessorBase):
    """
    Provides a bridge between the new v2 RecordProcessorBase, and the original RecordProcessorBase.

    This handles the conversion of the new input types to the older expected forms.  This normally shouldn't be used
    directly by record processors, since it's just a compatibility layer.

    The delegate should be a :py:class:`amazon_kclpy.kcl.RecordProcessorBase`:

    """

    def __init__(self, delegate):
        """
        Creates a new V2 to V3 record processor.

        :param amazon_kclpy.kcl.v2.RecordProcessorBase delegate: the delegate where requests will be forwarded to
        """
        self.delegate = delegate

    def initialize(self, initialize_input):
        """
        Initializes the record processor

        :param amazon_kclpy.messages.InitializeInput initialize_input: the initialization request
        :return: None
        """
        self.delegate.initialize(initialize_input)

    def process_records(self, process_records_input):
        """
        Expands the requests, and hands it off to the delegate for processing

        :param amazon_kclpy.messages.ProcessRecordsInput process_records_input: information about the records
            to process
        :return: None
        """
        self.delegate.process_records(process_records_input)

    def lease_lost(self, lease_lost_input):
        """
        Translates the lease lost call to the older shutdown/shutdown input style that was used.  In a special case the
        checkpointer will not be set in this call, which is essentially fine as checkpointing would fail anyway

        :param amazon_kclpy.messages.LeaseLostInput lease_lost_input: information about the lease loss
        (currently this is empty)
        :return: None
        """
        self.delegate.shutdown(ShutdownInput.zombie())

    def shard_ended(self, shard_ended_input):
        """
        Translates the shard end message to a shutdown input with a reason of TERMINATE and the checkpointer
        :param amazon_kclpy.messages.ShardEndedInput shard_ended_input: information, and checkpoint for the end of the
        shard.
        :return: None
        """
        self.delegate.shutdown(ShutdownInput.terminate(shard_ended_input.checkpointer))

    def shutdown_requested(self, shutdown_requested_input):
        """
        Sends the shutdown request to the delegate

        :param amazon_kclpy.messages.ShutdownRequested shutdown_requested_input: information related to the record processor shutdown
        :return: None
        """
        self.delegate.shutdown_requested(shutdown_requested_input)



#!/usr/bin/env python

# Copyright 2014-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from __future__ import print_function

import sys
import time

from amazon_kclpy import kcl
from amazon_kclpy.v3 import processor


class RecordProcessor(processor.RecordProcessorBase):
    """
    A RecordProcessor processes data from a shard in a stream. Its methods will be called with this pattern:

    * initialize will be called once
    * process_records will be called zero or more times
    * shutdown will be called if this MultiLangDaemon instance loses the lease to this shard, or the shard ends due
        a scaling change.
    """
    def __init__(self):
        self._SLEEP_SECONDS = 5
        self._CHECKPOINT_RETRIES = 5
        self._CHECKPOINT_FREQ_SECONDS = 60
        self._largest_seq = (None, None)
        self._largest_sub_seq = None
        self._last_checkpoint_time = None

    def log(self, message):
        sys.stderr.write(message)

    def initialize(self, initialize_input):
        """
        Called once by a KCLProcess before any calls to process_records

        :param amazon_kclpy.messages.InitializeInput initialize_input: Information about the lease that this record
            processor has been assigned.
        """
        self._largest_seq = (None, None)
        self._last_checkpoint_time = time.time()

    def checkpoint(self, checkpointer, sequence_number=None, sub_sequence_number=None):
        """
        Checkpoints with retries on retryable exceptions.

        :param amazon_kclpy.kcl.Checkpointer checkpointer: the checkpointer provided to either process_records
            or shutdown
        :param str or None sequence_number: the sequence number to checkpoint at.
        :param int or None sub_sequence_number: the sub sequence number to checkpoint at.
        """
        for n in range(0, self._CHECKPOINT_RETRIES):
            try:
                checkpointer.checkpoint(sequence_number, sub_sequence_number)
                return
            except kcl.CheckpointError as e:
                if 'ShutdownException' == e.value:
                    #
                    # A ShutdownException indicates that this record processor should be shutdown. This is due to
                    # some failover event, e.g. another MultiLangDaemon has taken the lease for this shard.
                    #
                    print('Encountered shutdown exception, skipping checkpoint')
                    return
                elif 'ThrottlingException' == e.value:
                    #
                    # A ThrottlingException indicates that one of our dependencies is is over burdened, e.g. too many
                    # dynamo writes. We will sleep temporarily to let it recover.
                    #
                    if self._CHECKPOINT_RETRIES - 1 == n:
                        sys.stderr.write('Failed to checkpoint after {n} attempts, giving up.\n'.format(n=n))
                        return
                    else:
                        print('Was throttled while checkpointing, will attempt again in {s} seconds'
                              .format(s=self._SLEEP_SECONDS))
                elif 'InvalidStateException' == e.value:
                    sys.stderr.write('MultiLangDaemon reported an invalid state while checkpointing.\n')
                else:  # Some other error
                    sys.stderr.write('Encountered an error while checkpointing, error was {e}.\n'.format(e=e))
            time.sleep(self._SLEEP_SECONDS)

    def process_record(self, data, partition_key, sequence_number, sub_sequence_number):
        """
        Called for each record that is passed to process_records.

        :param str data: The blob of data that was contained in the record.
        :param str partition_key: The key associated with this record.
        :param int sequence_number: The sequence number associated with this record.
        :param int sub_sequence_number: the sub sequence number associated with this record.
        """
        ####################################
        # Insert your processing logic here
        ####################################
        self.log("Record (Partition Key: {pk}, Sequence Number: {seq}, Subsequence Number: {sseq}, Data Size: {ds}"
                 .format(pk=partition_key, seq=sequence_number, sseq=sub_sequence_number, ds=len(data)))

    def should_update_sequence(self, sequence_number, sub_sequence_number):
        """
        Determines whether a new larger sequence number is available

        :param int sequence_number: the sequence number from the current record
        :param int sub_sequence_number: the sub sequence number from the current record
        :return boolean: true if the largest sequence should be updated, false otherwise
        """
        return self._largest_seq == (None, None) or sequence_number > self._largest_seq[0] or \
            (sequence_number == self._largest_seq[0] and sub_sequence_number > self._largest_seq[1])

    def process_records(self, process_records_input):
        """
        Called by a KCLProcess with a list of records to be processed and a checkpointer which accepts sequence numbers
        from the records to indicate where in the stream to checkpoint.

        :param amazon_kclpy.messages.ProcessRecordsInput process_records_input: the records, and metadata about the
            records.
        """
        try:
            for record in process_records_input.records:
                data = record.binary_data
                seq = int(record.sequence_number)
                sub_seq = record.sub_sequence_number
                key = record.partition_key
                self.process_record(data, key, seq, sub_seq)
                if self.should_update_sequence(seq, sub_seq):
                    self._largest_seq = (seq, sub_seq)

            #
            # Checkpoints every self._CHECKPOINT_FREQ_SECONDS seconds
            #
            if time.time() - self._last_checkpoint_time > self._CHECKPOINT_FREQ_SECONDS:
                self.checkpoint(process_records_input.checkpointer, str(self._largest_seq[0]), self._largest_seq[1])
                self._last_checkpoint_time = time.time()

        except Exception as e:
            self.log("Encountered an exception while processing records. Exception was {e}\n".format(e=e))

    def lease_lost(self, lease_lost_input):
        self.log("Lease has been lost")

    def shard_ended(self, shard_ended_input):
        self.log("Shard has ended checkpointing")
        shard_ended_input.checkpointer.checkpoint()

    def shutdown_requested(self, shutdown_requested_input):
        self.log("Shutdown has been requested, checkpointing.")
        shutdown_requested_input.checkpointer.checkpoint()


if __name__ == "__main__":
    kcl_process = kcl.KCLProcess(RecordProcessor())
    kcl_process.run()



'''
Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0


BEFORE YOU GET STARTED
======================

Before running the samples, you'll want to make sure that your environment is
configured to allow the samples to use your AWS credentials. To familiarize
yourself with AWS Credentials read this guide:

    http://docs.aws.amazon.com/general/latest/gr/aws-security-credentials.html

For the MultiLangDaemon and boto libs you'll want to make your credentials 
available to one of the credentials providers in the default credential
providers chain such as providing a ~/.aws/credentials file 

    http://docs.aws.amazon.com/AWSJavaSDK/latest/javadoc/com/amazonaws/auth/DefaultAWSCredentialsProviderChain.html

RUNNING THE SAMPLE
==================

Navigate to the amazon_kclpy directory and install the package. Using the amazon_kclpy
package requires the MultiLangDaemon which is provided by the java KCL. To get
the necessary jars to this directory before installing, you'll want to run the
"download_jars" command before running "install". If you just want the python
KCL and plan to retrieve the necessary jars yourself, you can just do "install"

    python setup.py download_jars
    python setup.py install

Now the amazon_kclpy and boto and required jars should be installed in your
environment. To start the sample putter, run:

    sample_kinesis_wordputter.py --stream words -w cat -w dog -w bird

This will create a Kinesis stream called words and put the words specified by
the -w options into the stream once each. Use -p SECONDS to indicate a period
over which to repeatedly put these words.

Now we would like to run a python KCL application that reads records from
the stream we just created, but first take a look in the samples directory,
you'll find a file called sample.properties, cat that file:

    cat samples/sample.properties

You'll see several properties defined there. "executableName" indicates the
executable for the MultiLangDaemon to run, "streamName" is the Kinesis stream
to read from, "appName" is the KCL application name to use which will be the
name of a DynamoDB table that gets created by the KCL, "initialPositionInStream"
tells the KCL how to start reading from shards upon a fresh startup. To run the
sample application you can use a helper script included in the package.

    amazon_kclpy_helper.py --print_command \
        --java <path-to-java> --properties samples/sample.properties

This will print the command needed to run the sample which you can copy paste,
or surround the command with back ticks, e.g.

    `amazon_kclpy_helper.py --print_command \
        --java <path-to-java> --properties samples/sample.properties`
'''



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
from mock import Mock
from amazon_kclpy import kcl, dispatch
from utils import make_io_obj


def build_basic_io_handler_mock(read_line_side_effects):
    """

    :param read_line_side_effects:
    :rtype: kcl._IOHandler
    """
    io_handler = Mock()
    io_handler.read_line.side_effect = read_line_side_effects
    io_handler.load_action.side_effect = lambda x: json.loads(x, object_hook=dispatch.message_decode)
    return io_handler


def test_checkpointer_exception():
    exception_name = 'ThisIsATestException'
    checkpointer = kcl.Checkpointer(
        build_basic_io_handler_mock(['{"action": "checkpoint",'
                                     '"checkpoint":"456", "sequenceNumber": "1234", "subSequenceNumber": 0, '
                                     '"error" : "' + exception_name + '"}']))
    try:
        checkpointer.checkpoint()
        assert 0, "Checkpointing should have raised an exception"
    except kcl.CheckpointError as e:
        assert e.value == exception_name


def test_checkpointer_unexpected_message_after_checkpointing():
    io_handler = Mock()
    io_handler.read_line.side_effect = ['{"action":"initialize", "shardId" : "shardid-123", '
                                        '"sequenceNumber": "1234", "subSequenceNumber": 1}', ]
    io_handler.load_action.side_effect = lambda x: json.loads(x, object_hook=dispatch.message_decode)
    checkpointer = kcl.Checkpointer(
        build_basic_io_handler_mock(
            ['{"action":"initialize", "shardId" : "shardid-123", "sequenceNumber": "1234", "subSequenceNumber": 1}']))

    try:
        checkpointer.checkpoint()
        assert 0, "Checkpointing should have raised an exception"
    except kcl.CheckpointError as e:
        assert e.value == 'InvalidStateException'


def test_kcl_process_exits_on_record_processor_exception():
    unique_string = "Super uniqe statement we can look for"
    errorFile = make_io_obj()
    class ClientException(Exception):
        pass
    mock_rp = Mock()  # type: kcl.RecordProcessorBase
    # Our record processor will just fail during initialization
    mock_rp.initialize.side_effect = [ClientException(unique_string)]
    kcl_process = kcl.KCLProcess(mock_rp,
                             input_file=make_io_obj('{"action":"initialize", "shardId" : "shardid-123", '
                                                   '"sequenceNumber": "1234", "subSequenceNumber": 1}'),
                             output_file=make_io_obj(),
                             error_file=errorFile)
    try:
        kcl_process.run()
    except ClientException:
        assert 0, "Should not have seen the ClientException propagate up the call stack."
    assert errorFile.getvalue().count(unique_string) > 0, 'We should see our error message printed to the error file'


def test_kcl_process_exits_on_action_message_exception():
    mock_rp = Mock()  # type: kcl.RecordProcessorBase
    # Our record processor will just fail during initialization
    kcl_process = kcl.KCLProcess(mock_rp,
                                 # This will suffice because a checkpoint message won't be understood by
                                 # the KCLProcessor (only the Checkpointer understands them)
                             input_file=make_io_obj('{"action":"invalid", "error" : "badstuff", '
                                                   '"sequenceNumber": "1234", "subSequenceNumber": 1}'),
                             output_file=make_io_obj(),
                             error_file=make_io_obj())
    try:
        kcl_process.run()
        assert 0, 'Should have received an exception here'
    except dispatch.MalformedAction:
        pass




# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This package provides an interface to the KCL MultiLangDaemon. This interface
manages the interaction with the MultiLangDaemon so that developers can focus
on implementing their record processor. A record processor executable typically
looks something like::
    #!env python
    from amazon_kclpy import kcl
    import json, base64

    class RecordProcessor(kcl.RecordProcessorBase):

        def initialize(self, shard_id):
            pass

        def process_records(self, records, checkpointer):
            pass

        def shutdown(self, checkpointer, reason):
            pass

    if __name__ == "__main__":
        kclprocess = kcl.KCLProcess(RecordProcessor())
        kclprocess.run()
"""



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

class CheckpointError(Exception):
    """
    Error class used for wrapping exception names passed through the input file.
    """
    def __init__(self, value):
        """
        :type value: str
        :param value: The name of the exception that was received while checkpointing. For more details see
            https://github.com/awslabs/amazon-kinesis-client/tree/master/src/main/java/com/amazonaws/services/kinesis/clientlibrary/exceptions
            Any of those exceptions' names could be returned by the MultiLangDaemon as a response to a checkpoint action.
        """
        self.value = value

    def __str__(self):
        return repr(self.value)




#!env python
'''
Copyright 2014-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
'''
from __future__ import print_function

import argparse
import sys
import time

import boto3

def get_stream_status(kinesis, stream_name):
    '''
    Query this provided connection object for the provided stream's status.
    :type conn: Kinesis.Client
    :param conn: A connection to Amazon Kinesis
    :type stream_name: str
    :param stream_name: The name of a stream.
    :rtype: str
    :return: The stream's status
    '''
    r = kinesis.describe_stream(StreamName=stream_name)
    description = r.get('StreamDescription')
    return description.get('StreamStatus')

def wait_for_stream(kinesis, stream_name):
    '''
    Wait for the provided stream to become active.
    :type kinesis: Kinesis.Client
    :param kinesis: A low-level client representing Amazon Kinesis
    :type stream_name: str
    :param stream_name: The name of a stream.
    '''
    SLEEP_TIME_SECONDS = 3
    status = get_stream_status(kinesis, stream_name)
    while status != 'ACTIVE':
        print('{stream_name} has status: {status}, sleeping for {secs} seconds'.format(
                stream_name = stream_name,
                status      = status,
                secs        = SLEEP_TIME_SECONDS))
        time.sleep(SLEEP_TIME_SECONDS) # sleep for 3 seconds
        status = get_stream_status(kinesis, stream_name)

def put_words_in_stream(kinesis, stream_name, words):
    '''
    Put each word in the provided list of words into the stream.
    :type kinesis: Kinesis.Client
    :param kinesis: A connection to Amazon Kinesis
    :type stream_name: str
    :param stream_name: The name of a stream.
    :type words: list
    :param words: A list of strings to put into the stream.
    '''
    for w in words:
        try:
            kinesis.put_record(StreamName=stream_name, Data=w, PartitionKey=w)
            print("Put word: " + w + " into stream: " + stream_name)
        except Exception as e:
            sys.stderr.write("Encountered an exception while trying to put a word: "
                             + w + " into stream: " + stream_name + " exception was: " + str(e))

def put_words_in_stream_periodically(conn, stream_name, words, period_seconds):
    '''
    Puts words into a stream, then waits for the period to elapse then puts the words in again. There is no strict
    guarantee about how frequently we put each word into the stream, just that we will wait between iterations.
    :type conn: boto.kinesis.layer1.KinesisConnection
    :param conn: A connection to Amazon Kinesis
    :type stream_name: str
    :param stream_name: The name of a stream.
    :type words: list
    :param words: A list of strings to put into the stream.
    :type period_seconds: int
    :param period_seconds: How long to wait, in seconds, between iterations over the list of words.
    '''
    while True:
        put_words_in_stream(conn, stream_name, words)
        print("Sleeping for {period_seconds} seconds".format(period_seconds=period_seconds))
        time.sleep(period_seconds)

if __name__ == '__main__':
    parser = argparse.ArgumentParser('''
Puts words into a stream.
# Using the -w option multiple times
sample_wordputter.py -s STREAM_NAME -w WORD1 -w WORD2 -w WORD3 -p 3
# Passing input from STDIN
echo "WORD1\\nWORD2\\nWORD3" | sample_wordputter.py -s STREAM_NAME -p 3
''')
    parser.add_argument("-s", "--stream", dest="stream_name", required=True,
                      help="The stream you'd like to create.", metavar="STREAM_NAME",)
    parser.add_argument("-r", "--regionName", "--region", dest="region", default="us-east-1",
                      help="The region you'd like to make this stream in. Default is 'us-east-1'", metavar="REGION_NAME",)
    parser.add_argument("-w", "--word", dest="words", default=[], action="append",
                      help="A word to add to the stream. Can be specified multiple times to add multiple words.", metavar="WORD",)
    parser.add_argument("-p", "--period", dest="period", type=int,
                      help="If you'd like to repeatedly put words into the stream, this option provides the period for putting "
                            + "words into the stream in SECONDS. If no period is given then the words are put once.",
                      metavar="SECONDS",)
    args = parser.parse_args()
    stream_name = args.stream_name

    '''
    Getting a connection to Amazon Kinesis will require that you have your credentials available to
    one of the standard credentials providers.
    '''
    print("Connecting to stream: {s} in {r}".format(s=stream_name, r=args.region))
    kinesis = boto3.client('kinesis', region_name=args.region)

    try:
        status = get_stream_status(kinesis, stream_name)
        if 'DELETING' == status:
            print('The stream: {s} is being deleted, please rerun the script.'.format(s=stream_name))
            sys.exit(1)
        elif 'ACTIVE' != status:
            wait_for_stream(kinesis, stream_name)
    except:
        # We'll assume the stream didn't exist so we will try to create it with just one shard
        kinesis.create_stream(StreamName=stream_name, ShardCount=1)
        wait_for_stream(kinesis, stream_name)
    # Now the stream should exist
    if len(args.words) == 0:
        print('No -w options provided. Waiting on input from STDIN')
        words = [l.strip() for l in sys.stdin.readlines() if l.strip() != '']
    else:
        words = args.words
    if args.period != None:
        put_words_in_stream_periodically(kinesis, stream_name, words, args.period)
    else:
        put_words_in_stream(kinesis, stream_name, words)


# Copyright 2014-2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from __future__ import print_function

import glob
import sys

import os
import shutil
import xml.etree.ElementTree as ET

from setuptools import Command
from setuptools import setup
from setuptools.command.install import install

if sys.version_info[0] >= 3:
    # Python 3
    from urllib.request import urlopen
else:
    # Python 2
    from urllib2 import urlopen

#
# This script modifies the basic setuptools by adding some functionality to the standard
# "install" command and by adding an additional command "download_jars" which
# simplifies retrieval of the jars required to run the KCL multi-language daemon
# which is required to run the sample app included in this package.
#
# If a user runs the basic install:
#
#     python setup.py install
#
# They will be notified of any jars that are downloaded for this package. Those jars
# will go in amazon_kclpy/jars so that they can be installed as part of this package's
# data.
#
#     python setup.py download_jars
#
# Will retrieve the configured jars from maven and then advise the user
# to rerun the install command.
#

PACKAGE_NAME = 'amazon_kclpy'
JAR_DIRECTORY = os.path.join(PACKAGE_NAME, 'jars')
PACKAGE_VERSION = '2.1.5'
PYTHON_REQUIREMENTS = [
    'boto3',
    # argparse is part of python2.7 but must be declared for python2.6
    'argparse',
]
REMOTE_MAVEN_PACKAGES_FILE = 'pom.xml'

class MavenJarDownloader:

    def __init__(self, on_completion, destdir=JAR_DIRECTORY, packages_file=REMOTE_MAVEN_PACKAGES_FILE):
        self.on_completion = on_completion
        self.destdir = destdir
        self.packages_file = packages_file
        self.packages = self.parse_packages_from_pom()

    def warning_string(self, missing_jars=[]):
        s = '''The following jars were not installed because they were not
present in this package at the time of installation:'''
        for jar in missing_jars:
            s += '\n  {jar}'.format(jar=jar)
        s += '''
This doesn't affect the rest of the installation, but may make it more
difficult for you to run the sample app and get started.

You should consider running:

    python setup.py download_jars
    python setup.py install

Which will download the required jars and rerun the install.
'''
        return s

    def parse_packages_from_pom(self):
        maven_root = ET.parse(self.packages_file).getroot()
        maven_version = '{http://maven.apache.org/POM/4.0.0}'
        # dictionary of common package versions encoded in `properties` section
        properties = {f"${{{child.tag.replace(maven_version, '')}}}": child.text
                      for child in maven_root.find(f'{maven_version}properties').iter() if 'version' in child.tag}

        packages = []
        for dep in maven_root.iter(f'{maven_version}dependency'):
            dependency = []
            for attr in ['groupId', 'artifactId', 'version']:
                val = dep.find(maven_version + attr).text
                if val in properties:
                    dependency.append(properties[val])
                else:
                    dependency.append(val)
            packages.append(tuple(dependency))

        return packages

    def download_and_check(self):
        self.download_files()
        self.on_completion()
        missing_jars = self.missing_jars()
        if len(missing_jars) > 0:
            raise RuntimeError(self.warning_string(missing_jars))

    def package_destination(self, artifact_id, version):
        return '{artifact_id}-{version}.jar'.format(artifact_id=artifact_id, version=version)

    def missing_jars(self):
        file_list = [os.path.join(self.destdir, self.package_destination(p[1], p[2])) for p in self.packages]
        return [f for f in file_list if not os.path.isfile(f)] # The missing files

    def package_url(self, group_id, artifact_id, version):
        #
        # Sample url:
        # https://search.maven.org/remotecontent?filepath=org/apache/httpcomponents/httpclient/4.2/httpclient-4.2.jar
        #
        prefix = 'https://search.maven.org/remotecontent?filepath='
        return '{prefix}{path}/{artifact_id}/{version}/{dest}'.format(
                                        prefix=prefix,
                                        path='/'.join(group_id.split('.')),
                                        artifact_id=artifact_id,
                                        version=version,
                                        dest=self.package_destination(artifact_id, version))

    def download_file(self, url, dest):
        """
        Downloads a file at the url to the destination.
        """
        print('Attempting to retrieve remote jar {url}'.format(url=url))
        try:
            response = urlopen(url)
            with open(dest, 'wb') as dest_file:
                shutil.copyfileobj(response, dest_file)
            print('Saving {url} -> {dest}'.format(url=url, dest=dest))
        except Exception as e:
            print('Failed to retrieve {url}: {e}'.format(url=url, e=e))
            return

    def download_files(self):
        for package in self.packages:
            dest = os.path.join(self.destdir, self.package_destination(package[1], package[2]))
            if os.path.isfile(dest):
                print('Skipping download of {dest}'.format(dest=dest))
            else:
                url = self.package_url(package[0], package[1], package[2])
                self.download_file(url, dest)


class DownloadJarsCommand(Command):
    description = "Download the jar files needed to run the sample application"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """
        Runs when this command is given to setup.py
        """
        downloader = MavenJarDownloader(on_completion=lambda : None)
        downloader.download_files()
        print('''
Now you should run:

    python setup.py install

Which will finish the installation.
''')


class InstallThenCheckForJars(install):

    def do_install(self):
        install.run(self)

    def run(self):
        """
        We override the basic install command. First we download jars then
        we run the basic install then we check whether the jars are present
        in this package. If they aren't present we warn the user and give
        them some advice on how to retry getting the jars.
        """
        downloader = MavenJarDownloader(self.do_install)
        downloader.download_and_check()


try:
    from wheel.bdist_wheel import bdist_wheel


    class BdistWheelWithJars(bdist_wheel):
        """
        This overrides the bdist_wheel command, that handles building a binary wheel of the package.
        Currently, as far as I can tell, binary wheel creation only occurs during the virtual environment creation.
        The package that bdist_wheel comes from isn't a modeled dependency of this package, but is required for virtual
        environment creation.
        """

        def do_run(self):
            bdist_wheel.run(self)

        def run(self):
            downloader = MavenJarDownloader(self.do_run)
            downloader.download_and_check()

except ImportError:
    pass

if __name__ == '__main__':
    commands = {
        'download_jars': DownloadJarsCommand,
        'install': InstallThenCheckForJars,
    }
    try:
        #
        # BdistWheelWithJars will only be present if the wheel package is present, and that is present during
        # virtual environment creation.
        # It's important to note this is a hack.  There doesn't appear to be a way to execute hooks around wheel
        # creation by design.  See https://github.com/pypa/packaging-problems/issues/64 for more information.
        #
        commands['bdist_wheel'] = BdistWheelWithJars
    except NameError:
        pass

    setup(
        name=PACKAGE_NAME + "_ekirmayer",
        version=PACKAGE_VERSION,
        description='A python interface for the Amazon Kinesis Client Library MultiLangDaemon',
        license='Apache-2.0',
        packages=[PACKAGE_NAME, PACKAGE_NAME + "/v2", PACKAGE_NAME + "/v3", 'samples'],
        scripts=glob.glob('samples/*py'),
        package_data={
            '': ['*.txt', '*.md'],
            PACKAGE_NAME: ['jars/*'],
            'samples': ['sample.properties'],
        },
        install_requires=PYTHON_REQUIREMENTS,
        setup_requires=["pytest-runner"],
        tests_require=["pytest", "mock"],
        cmdclass=commands,
        url="https://github.com/awslabs/amazon-kinesis-client-python",
        keywords="amazon kinesis client library python",
        zip_safe=False,
        )



import json
import re

from amazon_kclpy import kcl
from utils import make_io_obj


# Dummy record processor
class RecordProcessor(kcl.RecordProcessorBase):

    def __init__(self, expected_shard_id, expected_sequence_number):
        self.expected_shard_id = expected_shard_id
        self.expected_sequence_number = expected_sequence_number
        pass

    def initialize(self, shard_id):
        assert shard_id == self.expected_shard_id
        pass

    def process_records(self, records, checkpointer):
        seq = records[0].get('sequenceNumber')
        assert seq == self.expected_sequence_number
        try:
            checkpointer.checkpoint(seq)
            assert 0, "First checkpoint should fail"
        except Exception:
            # Try it one more time (this time it'll work)
            checkpointer.checkpoint(seq)

    def shutdown(self, checkpointer, reason):
        if 'TERMINATE' == reason:
            checkpointer.checkpoint()


'''
An input string which we'll feed to a file for kcl.py to read from.
'''

'''
This string is approximately what the output should look like. We remove whitespace when comparing this to what is
written to the outputfile.
'''
test_output_string = """
{"action": "status", "responseFor": "initialize"}
{"action": "checkpoint", "checkpoint": "456"}
{"action": "checkpoint", "checkpoint": "456"}
{"action": "status", "responseFor": "processRecords"}
{"action": "checkpoint", "checkpoint": null}
{"action": "status", "responseFor": "shutdown"}
"""

test_output_messages = [
    {"action": "status", "responseFor": "initialize"},
    {"action": "checkpoint", "sequenceNumber": "456", "subSequenceNumber": None},
    {"action": "checkpoint", "sequenceNumber": "456", "subSequenceNumber": None},
    {"action": "status", "responseFor": "processRecords"},
    {"action": "checkpoint", "sequenceNumber": None, "subSequenceNumber": None},
    {"action": "status", "responseFor": "shardEnded"}
]


def _strip_all_whitespace(s):
    return re.sub('\\s*', '', s)


test_shard_id = "shardId-123"
test_sequence_number = "456"

test_input_messages = [
    {"action": "initialize", "shardId": test_shard_id, "sequenceNumber": test_sequence_number, "subSequenceNumber": 0},
    {"action": "processRecords", "millisBehindLatest": 1476889708000, "records":
        [
            {
                "action": "record", "data": "bWVvdw==", "partitionKey": "cat", "sequenceNumber": test_sequence_number,
                "subSequenceNumber": 0, "approximateArrivalTimestamp": 1476889707000
            }
        ]
     },
    {"action": "checkpoint", "sequenceNumber": test_sequence_number, "subSequenceNumber": 0, "error": "Exception"},
    {"action": "checkpoint", "sequenceNumber": test_sequence_number, "subSequenceNumber": 0},
    {"action": "shardEnded"},
    {"action": "checkpoint", "sequenceNumber": test_sequence_number, "subSequenceNumber": 0}
]


def test_kcl_py_integration_test_perfect_input():
    test_input_json = "\n".join(map(lambda j: json.dumps(j), test_input_messages))
    input_file = make_io_obj(test_input_json)
    output_file = make_io_obj()
    error_file = make_io_obj()
    process = kcl.KCLProcess(RecordProcessor(test_shard_id, test_sequence_number),
                             input_file=input_file, output_file=output_file, error_file=error_file)
    process.run()
    '''
    The strings are approximately the same, modulo whitespace.
    '''
    output_message_list = filter(lambda s: s != "", output_file.getvalue().split("\n"))
    responses = [json.loads(s) for s in output_message_list]
    assert len(responses) == len(test_output_messages)
    for i in range(len(responses)):
        assert responses[i] == test_output_messages[i]

    '''
    There should be some error output but it seems like overly specific to make sure that a particular message is printed.
    '''
    error_output = error_file.getvalue()
    assert error_output == ""



# Copyright 2014-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
import json
import sys
import traceback

from amazon_kclpy import dispatch
from amazon_kclpy.v2 import processor as v2processor
from amazon_kclpy.v3 import processor as v3processor
from amazon_kclpy import messages
from amazon_kclpy.checkpoint_error import CheckpointError


class _IOHandler(object):
    """
    Hidden class used by :class:`KCLProcess` and :class:`Checkpointer` to communicate with the input and output
    files.
    """

    def __init__(self, input_file, output_file, error_file):
        """
        :param file input_file: A file to read input lines from (e.g. sys.stdin).
        :param file output_file: A file to write output lines to (e.g. sys.stdout).
        :param file error_file: A file to write error lines to (e.g. sys.stderr).
        """
        self.input_file = input_file
        self.output_file = output_file
        self.error_file = error_file

    def write_line(self, line):
        """
        Writes a line to the output file. The line is preceded and followed by a new line because other libraries
        could be writing to the output file as well (e.g. some libs might write debugging info to STDOUT) so we would
        like to prevent our lines from being interlaced with other messages so the MultiLangDaemon can understand them.

        :param str line: A line to write (e.g. '{"action" : "status", "responseFor" : "<someAction>"}')
        """
        self.output_file.write('\n{line}\n'.format(line=line))
        self.output_file.flush()

    def write_error(self, error_message):
        """
        Write a line to the error file.

        :param str error_message: An error message.
        """
        self.error_file.write('{error_message}\n'.format(error_message=error_message))
        self.error_file.flush()

    def read_line(self):
        """
        Reads a line from the input file.

        :rtype: str
        :return: A single line read from the input_file (e.g. '{"action" : "initialize", "shardId" : "shardId-000001"}')
        """
        return self.input_file.readline()

    def load_action(self, line):
        """
        Decodes a message from the MultiLangDaemon.
        :type line: str
        :param line: A message line that was delivered received from the MultiLangDaemon (e.g.
            '{"action" : "initialize", "shardId" : "shardId-000001"}')

        :rtype: amazon_kclpy.messages.MessageDispatcher
        :return: A callable action class that contains the action presented in the line
        """
        return json.loads(line, object_hook=dispatch.message_decode)

    def write_action(self, response):
        """
        :type response: dict
        :param response: A dictionary with an action message such as 'checkpoint' or 'status'. For example if the action that was
            just handled by this processor was an 'initialize' action, this dictionary would look like
            {'action' : status', 'responseFor' : 'initialize'}
        """
        self.write_line(json.dumps(response))


CheckpointError = CheckpointError


class Checkpointer(object):
    """
    A checkpointer class which allows you to make checkpoint requests. A checkpoint marks a point in a shard
    where you've successfully processed to. If this processor fails or loses its lease to that shard, another
    processor will be started either by this MultiLangDaemon or a different instance and resume at the most recent
    checkpoint in this shard.
    """
    def __init__(self, io_handler):
        """
        :type io_handler: amazon_kclpy.kcl._IOHandler
        :param io_handler: An IOHandler object which this checkpointer will use to write and read checkpoint actions
            to and from the MultiLangDaemon.
        """
        self.io_handler = io_handler

    def _get_action(self):
        """
        Gets the next json message from STDIN

        :rtype: object
        :return: Either a child of MessageDispatcher, or a housekeeping object type
        """
        line = self.io_handler.read_line()
        action = self.io_handler.load_action(line)
        return action

    def checkpoint(self, sequence_number=None, sub_sequence_number=None):
        """
        Checkpoints at a particular sequence number you provide or if no sequence number is given, the checkpoint will
        be at the end of the most recently delivered list of records

        :param str or None sequence_number: The sequence number to checkpoint at or None if you want to checkpoint at the
            farthest record
        :param int or None sub_sequence_number: the sub sequence to checkpoint at, if set to None will checkpoint
            at the farthest sub_sequence_number
        """
        response = {"action": "checkpoint", "sequenceNumber": sequence_number, "subSequenceNumber": sub_sequence_number}
        self.io_handler.write_action(response)
        action = self._get_action()
        if isinstance(action, messages.CheckpointInput):
            if action.error is not None:
                raise CheckpointError(action.error)
        else:
            #
            # We are in an invalid state. We will raise a checkpoint exception
            # to the RecordProcessor indicating that the KCL (or KCLpy) is in
            # an invalid state. See KCL documentation for description of this
            # exception. Note that the documented guidance is that this exception
            # is NOT retryable so the client code should exit.
            #
            raise CheckpointError('InvalidStateException')


# RecordProcessor base class
class RecordProcessorBase(object):
    """
    Base class for implementing a record processor.A RecordProcessor processes a shard in a stream.
    Its methods will be called with this pattern:

    - initialize will be called once
    - process_records will be called zero or more times
    - shutdown will be called if this MultiLangDaemon instance loses the lease to this shard
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def initialize(self, shard_id):
        """
        Called once by a KCLProcess before any calls to process_records

        :type shard_id: str
        :param shard_id: The shard id that this processor is going to be working on.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def process_records(self, records, checkpointer):
        """
        Called by a KCLProcess with a list of records to be processed and a checkpointer which accepts sequence numbers
        from the records to indicate where in the stream to checkpoint.

        :type records: list
        :param records: A list of records that are to be processed. A record looks like
            {"data":"<base64 encoded string>","partitionKey":"someKey","sequenceNumber":"1234567890"} Note that "data" is a base64
            encoded string. You can use base64.b64decode to decode the data into a string. We currently do not do this decoding for you
            so as to leave it to your discretion whether you need to decode this particular piece of data.

        :type checkpointer: amazon_kclpy.kcl.Checkpointer
        :param checkpointer: A checkpointer which accepts a sequence number or no parameters.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown(self, checkpointer, reason):
        """
        Called by a KCLProcess instance to indicate that this record processor should shutdown. After this is called,
        there will be no more calls to any other methods of this record processor.

        :type checkpointer: amazon_kclpy.kcl.Checkpointer
        :param checkpointer: A checkpointer which accepts a sequence number or no parameters.

        :type reason: str
        :param reason: The reason this record processor is being shutdown, either TERMINATE or ZOMBIE. If ZOMBIE,
            clients should not checkpoint because there is possibly another record processor which has acquired the lease
            for this shard. If TERMINATE then checkpointer.checkpoint() should be called to checkpoint at the end of the
            shard so that this processor will be shutdown and new processor(s) will be created to for the child(ren) of
            this shard.
        """
        raise NotImplementedError

    def shutdown_requested(self, checkpointer):
        """
        Called by a KCLProcess instance to indicate that this record processor is about to be be shutdown.  This gives
        the record processor a chance to checkpoint, before the lease is terminated.

        :type checkpointer: amazon_kclpy.kcl.Checkpointer
        :param checkpointer: A checkpointer which accepts a sequence number or no parameters.
        """
        pass

    version = 1


class KCLProcess(object):

    def __init__(self, record_processor, input_file=sys.stdin, output_file=sys.stdout, error_file=sys.stderr):
        """
        :type record_processor: RecordProcessorBase or amazon_kclpy.v2.processor.RecordProcessorBase
        :param record_processor: A record processor to use for processing a shard.

        :param file input_file: A file to read action messages from. Typically STDIN.

        :param file output_file: A file to write action messages to. Typically STDOUT.

        :param file error_file: A file to write error messages to. Typically STDERR.
        """
        self.io_handler = _IOHandler(input_file, output_file, error_file)
        self.checkpointer = Checkpointer(self.io_handler)
        if record_processor.version == 2:
            self.processor = v3processor.V2toV3Processor(record_processor)
        elif record_processor.version == 1:
            self.processor = v3processor.V2toV3Processor(v2processor.V1toV2Processor(record_processor))
        else:
            self.processor = record_processor

    def _perform_action(self, action):
        """
        Maps input action to the appropriate method of the record processor.

        :type action:
        :param MessageDispatcher action: A derivative of MessageDispatcher that will handle the provided input

        :raises MalformedAction: Raised if the action is missing attributes.
        """

        try:
            action.dispatch(self.checkpointer, self.processor)
        except SystemExit as sys_exit:
            # On a system exit exception just go ahead and exit
            raise sys_exit
        except Exception as ex:
            """
            We don't know what the client's code could raise and we have no way to recover if we let it propagate
            up further. We will mimic the KCL and pass over client errors. We print their stack trace to STDERR to
            help them notice and debug this type of issue.
            """
            self.io_handler.error_file.write("Caught exception from action dispatch: {ex}".format(ex=str(ex)))
            traceback.print_exc(file=self.io_handler.error_file)
            self.io_handler.error_file.flush()

    def _report_done(self, response_for=None):
        """
        Writes a status message to the output file.

        :param response_for: Required parameter; the action that this status message is confirming completed.
        """
        self.io_handler.write_action({"action": "status", "responseFor": response_for})

    def _handle_a_line(self, line):
        """
        - Parses the line from JSON
        - Invokes the appropriate method of the RecordProcessor
        - Writes a status message back to MultiLanguageDaemon

        :type line: str
        :param line: A line that has been read from STDIN and is expected to be a JSON encoded dictionary
            representing what action to take.
        """
        action = self.io_handler.load_action(line)
        self._perform_action(action)
        self._report_done(action.action)

    def run(self):
        """
        Starts this KCL processor's main loop.
        """
        line = True
        """
        We don't make any effort to stop errors from propagating up and exiting the program
        because there is really nothing the KCLpy can do to recover after most exceptions that could
        occur, e.g. I/O error or json decoding exception (the MultiLangDaemon should never write a non-json string
        to this process).
        """
        while line:
            line = self.io_handler.read_line()
            if line:
                self._handle_a_line(line)


